#include "SSSP.h"
#include <array>
#include <vector>

void SSSP::setOpenmpConfig(int dynamic, int threads) {
  omp_set_dynamic(dynamic);
  omp_set_num_threads(threads);
}

void SSSP::runPullWithOpenMP(TradArrayNoDegEWGraphR* G_csr, int start_vertex, int max, int** sp) {
#ifdef __XGRAPH_COUNT_OPERATIONS__
  MemoryStats s_global;
  PapiStats papi_s_global(papi_set);
#endif

  v_id n = G_csr->n_;
  int* bucket = new int[n];

  int max_weight = max;

#pragma omp parallel for
  for (int i = 0; i < n; i++) {
    bucket[i] = max_weight;
  }

  bucket[start_vertex] = 0;

  int wg = 0;
  int iter = 1;

#pragma omp parallel 
  {
#ifdef __XGRAPH_COUNT_OPERATIONS__
    MemoryStats s;
    PapiStats papi_s(papi_set);
#endif

    while(wg < max_weight) {
#pragma omp single
      {
        cout << "Itr: " << iter << ", wg: " << wg << ", max wg: " << max_weight << endl;
      }

#ifdef __BENCHMARK__
#pragma omp single
      { 
        LSB_Set_Rparam_int("itr", iter);
        LSB_Set_Rparam_string("phs", "comp");
        LSB_Res();
      }
#endif

#ifdef __XGRAPH_COUNT_OPERATIONS__
      papi_s.start_papi_daint_counters();
#endif

#pragma omp for
      for (int i = 0; i < n; ++i) {
        int weight;
        if(bucket[i] > wg) {
          for (int e_id = 0; e_id < G_csr->d(i); ++e_id) {
            weight = bucket[G_csr->e(i, e_id)] + G_csr->w(i, e_id);
#pragma omp critical
            {
              if (weight < bucket[i]) {
                bucket[i] = weight;
              }
            }
#ifdef __XGRAPH_COUNT_OPERATIONS__
            ++s.omp_critical_;
#endif
          }
        }
      }

#ifdef __XGRAPH_COUNT_OPERATIONS__
      papi_s.stop_papi_daint_counters();
      PapiStats::increase_global_stats(&papi_s_global, &papi_s);

      MemoryStats::increase_lock_global_stats(&s_global, &s);
      s.omp_critical_ = 0;
#endif

#ifdef __BENCHMARK__
#pragma omp barrier
#pragma omp single
      {
#ifdef __XGRAPH_COUNT_OPERATIONS__
        s_global.record_lock_stats_in_lsb();
        papi_s_global.record_papi_in_lsb();
#endif

        LSB_Rec(run_id);
      }
#endif

#pragma omp single
      {
        ++wg;
        ++iter;
      }
    }
  }

  *sp = bucket;
}

void SSSP::runPushWithOpenMP(TradArrayNoDegEWGraphR* G_csr, int start_vertex, int max, int** sp) {
#ifdef __XGRAPH_COUNT_OPERATIONS__
  MemoryStats s_global;
  PapiStats papi_s_global(papi_set);
#endif

  v_id n = G_csr->n_;
  int* bucket = new int[n];

  int max_weight = max;

#pragma omp parallel for
  for (int i = 0; i < n; i++) {
    bucket[i] = max_weight;
  }

  vector<int> *buck = new vector<int>();
  vector<vector<int>> *bucket_list = new vector<vector<int>>();

  buck->push_back(start_vertex);
  bucket_list->insert(bucket_list->end(), *buck);

  unsigned int i = 0;
  int iter = 1;

#pragma omp parallel 
  {
#ifdef __XGRAPH_COUNT_OPERATIONS__
    MemoryStats s;
    PapiStats papi_s(papi_set);
#endif

    while(i < bucket_list->size()) {
#pragma omp single
      {
        cout << "Itr: " << iter << ", i: " << i << ", bucket list size: " << bucket_list->size() << endl;
      }

#ifdef __BENCHMARK__
#pragma omp single
      { 
        LSB_Set_Rparam_int("itr", iter);
        LSB_Set_Rparam_string("phs", "comp");
        LSB_Res();
      }
#endif

#ifdef __XGRAPH_COUNT_OPERATIONS__
      papi_s.start_papi_daint_counters();
#endif

#pragma omp for
      for (unsigned it = 0; it < bucket_list->at(i).size(); ++it) {
        int weight = 0;
        unsigned j = bucket_list->at(i).at(it);
        for (int e_id = 0; e_id < G_csr->d(j); ++e_id) {
          weight = i + G_csr->w(j, e_id);
          if (weight < bucket[G_csr->e(j, e_id)]) {
#pragma omp critical
            {
              if (weight >= (v_id)bucket_list->size()) {
                bucket_list->resize(weight+1);
              }
              bucket_list->at(weight).push_back(G_csr->e(j, e_id));
              bucket[G_csr->e(j, e_id)] = weight;
            }
#ifdef __XGRAPH_COUNT_OPERATIONS__
            ++s.omp_critical_;
#endif
          }
        }
      }

#ifdef __XGRAPH_COUNT_OPERATIONS__
      papi_s.stop_papi_daint_counters();
      PapiStats::increase_global_stats(&papi_s_global, &papi_s);

      MemoryStats::increase_lock_global_stats(&s_global, &s);
      s.omp_critical_ = 0;
#endif

#ifdef __BENCHMARK__
#pragma omp barrier
#pragma omp single
      {
#ifdef __XGRAPH_COUNT_OPERATIONS__
        s_global.record_lock_stats_in_lsb();
        papi_s_global.record_papi_in_lsb();
#endif

        LSB_Rec(run_id);
      }
#endif

#pragma omp single
      { 
        ++i;
        ++iter;
      }
    }
  }
  *sp = bucket;
}

/* It is here because logic and tools are similar to SSSP */
void SSSP::graphColoring(TradArrayNoDegGraphR* G_csr, int** sp, int maxiter) {
#ifdef __XGRAPH_COUNT_OPERATIONS__
  MemoryStats s_global;
  PapiStats papi_s_global(papi_set);
#endif

  v_id n = G_csr->n_;
  int* bucket = new int[n]();

  v_id* offsets = G_csr->offsets_;
  v_id* adj = G_csr->adj_;

#pragma omp parallel for
  for (v_id id = 0; id < n; ++id) {

    v_id id_offset = offsets[id];
    v_id neighs_v_nr = offsets[id + 1] - offsets[id];

    bool check = true;
    for (v_id i = 0; i < neighs_v_nr; ++i) {
      if (bucket[adj[id_offset + i]] != 0) {
        check = false;
      }
    }
    if (check) bucket[id] = 1;
  }

  int wg = 1;
  int iter = 1;

#pragma omp parallel 
  {
#ifdef __XGRAPH_COUNT_OPERATIONS__
    MemoryStats s;
    PapiStats papi_s(papi_set);
#endif

    while(wg < maxiter) {
#pragma omp single
      {
        cout << "Itr: " << iter << ", wg: " << wg << ", max iter: " << maxiter << endl;
      }

#ifdef __BENCHMARK__
#pragma omp single
      { 
        LSB_Set_Rparam_int("itr", iter);
        LSB_Set_Rparam_string("phs", "comp");
        LSB_Res();
      }
#endif

#ifdef __XGRAPH_COUNT_OPERATIONS__
      papi_s.start_papi_daint_counters();
#endif

#pragma omp for
      for (int i = 0; i < n; ++i) {
        if (bucket[i] == wg) {

          v_id id_offset = offsets[i];
          v_id neighs_v_nr = offsets[i + 1] - offsets[i];

          for (v_id id = 0; id < neighs_v_nr; ++id) {
#pragma omp critical
            {
              if (bucket[adj[id_offset + id]] == 0 || (bucket[i] == wg && bucket[adj[id_offset + id]] == wg)) {
                bucket[adj[id_offset + id]] = wg + 1;
              }
            }
#ifdef __XGRAPH_COUNT_OPERATIONS__
            ++s.omp_critical_;
#endif
          }
        }
      }

#ifdef __XGRAPH_COUNT_OPERATIONS__
      papi_s.stop_papi_daint_counters();
      PapiStats::increase_global_stats(&papi_s_global, &papi_s);

      MemoryStats::increase_lock_global_stats(&s_global, &s);
      s.omp_critical_ = 0;
#endif

#ifdef __BENCHMARK__
#pragma omp barrier
#pragma omp single
      {
#ifdef __XGRAPH_COUNT_OPERATIONS__
        s_global.record_lock_stats_in_lsb();
        papi_s_global.record_papi_in_lsb();
#endif

        LSB_Rec(run_id);
      }
#endif

#pragma omp single
      {
        ++wg;
      }
    }
  }

  *sp = bucket;
}
